#!/usr/bin/perl

###############################################################################
#
# FileLister - an open source alternative for listing files/directories
#              with an easy navigateable interface.
#
# Copyright (c) by  Oliver Pitzeier, June 2003
#                   oliver@linux-kernel.at
#
# Changes are welcome, but please inform me about those!
#
###############################################################################

###############################################################################
#
# $Id: main.pl,v 1.39 2004/01/19 12:48:44 oliver Exp $
#
# $Log: main.pl,v $
# Revision 1.39  2004/01/19 12:48:44  oliver
# * A lot of stuff was done:
#  + Add time_delta function - calculates the time delta between to given timestamps
#  + Add is_old function - returns 0 or 1 - uses delta to say file is old or not: used for filelist to check if new-flash.gif shall be displayed and for caching files
#
# Revision 1.38  2004/01/19 11:39:45  oliver
# + Create missing cache dir option added to config and well, it has been coded in main.pl as well. :-)
# + Fix a lot of identation since, caching makes a big new if. :-)
#
# Revision 1.37  2004/01/16 19:07:11  oliver
# + Bugfix: up.gif was hardcoded
#
# Revision 1.36  2004/01/15 18:04:11  oliver
# + Added caching stuff...
#
# Revision 1.35  2004/01/12 16:27:02  oliver
# Typo...
#
# Revision 1.34  2004/01/12 16:25:02  oliver
# * Make all pages XHTML 1.0 valid
#
# Revision 1.33  2004/01/12 13:18:47  oliver
# + Add option in configfile and code to enable/disable search possibility.
#
# Revision 1.32  2004/01/12 13:10:00  oliver
# + Add stuff for search and ToolTips
#
# Revision 1.31  2004/01/08 10:35:25  oliver
# + One more typo in copyright statement
#
# Revision 1.30  2004/01/08 10:34:37  oliver
# + Change copyright date.
#
# Revision 1.29  2004/01/08 09:49:43  oliver
# +/- Typo...
#
# Revision 1.28  2003/11/26 12:02:04  oliver
# + Added code to show User/Group information on directories
# + Added defaults for this in config-file
# + Added code to the templates to show the new information
#
# Revision 1.27  2003/10/16 13:33:51  oliver
# + Fix not working 'One up'
#
# Revision 1.26  2003/10/16 13:18:55  oliver
# +) removed: absolute URLs (to avoid problems with e.g. mod_proxy)
#
# Revision 1.25  2003/10/15 13:59:00  oliver
# + Short comments about the user/group stuff...
#
# Revision 1.24  2003/10/15 13:56:16  oliver
# + Mailto-stuff added
# + User/Group columns added
#
# Revision 1.23  2003/10/03 10:43:36  oliver
# + Add inline documentation
#
# Revision 1.22  2003/10/01 14:13:32  oliver
# + Make title unique for every host...
#
# Revision 1.21  2003/10/01 14:02:19  oliver
# + Fix warning at regex... Simply wrong coded...
#
# Revision 1.20  2003/09/30 16:17:25  oliver
# + Fix path to rpminfo - we should include this in config-file....
#
# Revision 1.19  2003/09/30 13:09:10  oliver
# + Fix some link-stuff; Only happens if you have DirectoryIndex main.pl; We do have this problem. :(
#
# Revision 1.18  2003/09/30 12:40:34  oliver
# + Two new options: enable-ftp-download and enable-http-download; You can define via the config-file, if you want your users to have the 'HTTP' and/or 'FTP' link
# + Note: If you update main.pl, you will have to enable one of those options in FileLister.yaml.conf, else you will have _NO_ download option...
#
# Revision 1.17  2003/09/30 11:20:32  oliver
# + Fix title...
#
# Revision 1.16  2003/09/30 11:20:13  oliver
# + More cosmetic stuff
#
# Revision 1.15  2003/09/26 14:08:41  oliver
# + More subfolder stuff...
#
# Revision 1.14  2003/09/26 13:24:07  oliver
# +BUG: back to main.tpl
#
# Revision 1.13  2003/09/26 13:23:41  oliver
# Some experimental stuff => RSS
#
# Revision 1.12  2003/09/26 12:33:48  oliver
# + Fix problem with regex in config file
#
# Revision 1.11  2003/09/26 11:47:50  oliver
# + Get hide_files from config-file...
#
# Revision 1.10  2003/09/26 11:45:57  oliver
# - One more bug... http_dlurl instead of http_url...
#
# Revision 1.9  2003/09/26 11:43:47  oliver
# + Buggy... Testing stuff included... :(
#
# Revision 1.8  2003/09/26 11:30:21  oliver
# + Switched configuration... We now have a YAML style config-file.
#
# Revision 1.7  2003/09/26 10:48:12  oliver
# + lettersurl is no longer in use...
#
# Revision 1.6  2003/09/26 10:14:10  oliver
# + One more stylesheet fix. I used the wrong one...
#
# Revision 1.5  2003/09/26 10:10:28  oliver
# + Stylesheet fixes...
#
# Revision 1.4  2003/09/26 09:39:15  oliver
# + Send VERSION to the templates...
#
# Revision 1.3  2003/09/26 09:38:21  oliver
# + Add the VERSION variable...
#
# Revision 1.2  2003/09/26 09:34:11  oliver
# + Change the creation date... WOW\! We work on it since June... :)
#
# Revision 1.1  2003/09/26 09:28:07  oliver
# + This it the new FileLister Application, based on Template::Toolkit
#
###############################################################################

# Load modules we use.
use strict;
use warnings;
require Template;
use CGI qw/:standard/;
use YAML::LoadFileCached;
use Storable;
use Date::Calc qw/Delta_YMDHMS/;
use File::Path;

sub time_delta($) {
        my $filename = shift;
        # Check the time-delta
        my ($sec,     $min,   $hour,   $mday,   $mon,   $year,   $wday,   $yday,   $isdst) =
            localtime(((stat("$filename")))[9]);
        my ($c_sec, $c_min, $c_hour, $c_mday, $c_mon, $c_year, $c_wday, $c_yday, $c_isdst) = localtime(time);
        $year += 1900; $c_year += 1900;
        $mon += 1; $c_mon += 1;
       
        my $delta;
        ($delta->{years}, $delta->{months}, $delta->{days}, $delta->{hours}, $delta->{minutes}, $delta->{seconds}) =
             Delta_YMDHMS(  $year,   $mon,   $mday,   $hour,   $min,   $sec,
                          $c_year, $c_mon, $c_mday, $c_hour, $c_min, $c_sec);
        return $delta;
}

sub is_old($$) {
        my $delta    = shift;
        my $timespec = shift;

        $timespec =~ m/(.*)(.)$/;
        my($count, $type) = ($1, $2);

        my $is_old = 0;

        # Time can be specified using 'h' for hours, 'm' for minutes and 's' for seconds.
        # Check for this here...
        if   ($type eq 'Y') { $is_old = 1 if($delta->{years}   > $count); }

        elsif($type eq 'M') { $is_old = 1 if($delta->{months}  > $count);
                              $is_old = 1 if($delta->{years}   >= 1);      }

        elsif($type eq 'D') { $is_old = 1 if($delta->{days}    > $count);
                              $is_old = 1 if($delta->{months}  >= 1);
                              $is_old = 1 if($delta->{years}   >= 1);      }

        elsif($type eq 'h') { $is_old = 1 if($delta->{hours}   > $count);
                              $is_old = 1 if($delta->{days}    >= 1);
                              $is_old = 1 if($delta->{months}  >= 1);
                              $is_old = 1 if($delta->{years}   >= 1);      }

        elsif($type eq 'm') { $is_old = 1 if($delta->{minutes} > $count);
                              $is_old = 1 if($delta->{hours}   >= 1);
                              $is_old = 1 if($delta->{days}    >= 1);
                              $is_old = 1 if($delta->{months}  >= 1);
                              $is_old = 1 if($delta->{years}   >= 1);      }

        elsif($type eq 's') { $is_old = 1 if($delta->{seconds} > $count);
                              $is_old = 1 if($delta->{minutes} >= 1);
                              $is_old = 1 if($delta->{hours}   >= 1);
                              $is_old = 1 if($delta->{days}    >= 1);
                              $is_old = 1 if($delta->{months}  >= 1);
                              $is_old = 1 if($delta->{years}   >= 1);      }

        else                { warn "Unknown 'timetype': $type, unknown (possible values: \"Y, M, D, h, m, s\")!"; $is_old = 1; }
        # This was time-delta stuff.

        return $is_old;
}

# Set Version - comes from CVS.
(our $VERSION) = '$Revision: 1.39 $' =~ /([\d.]+)/;

# Create a new TT2 object.
my $tt = Template->new;

# Set static variable configfile and load it using YAML.
my $configfile = 'config/FileLister.yaml.conf';
my $config = LoadFileCached($configfile);

# Set variable hide_files - comes from config file.
my $hide_files = $config->{'listing_stuff'}->{'hide_files'};

# Some security checks - double dots are simply ignored and if no path is
# specified, set it to "/" which is not the partition "/", but our root
# releative to the path from the config file.
my $current = param('current') || '/';
$current =~ s/\.\.//g;

# Some helper variables...
my $path_line;
my $oneup_url;
my @dir_list;
my @file_list;
my $main_data;
my $cached;

# Check to see if caching is enabled.
if($config->{'caching'}->{'enabled'}) {
    # Helper.
    my $store_filename = "FileLister".$current;
    $store_filename =~ s/\/$/$config->{'caching'}->{'suffix'}/g;
    $store_filename =~ s/\//_/g;

    # If the cache-file for this path exists.
    if(-f "$config->{'caching'}->{'path'}/$store_filename") {
        # Retrieve the cache from the file
        $main_data = retrieve("$config->{'caching'}->{'path'}/$store_filename");
        $cached = 1;
        warn "Using cached version from $config->{'caching'}->{'path'}/$store_filename" if $config->{'caching'}->{'debug'};
     }
}

# If main_data has been initialized and filled with data, we have choosen to use the cached version.
unless($main_data) {
    # Now that we can be sure we have a path and initialized our helpers, we
    # can open the directory and read the file-list.
    if(opendir(DIR, $config->{'path_and_url_stuff'}->{'path'}.$current)) {
        my @files = readdir(DIR);

        # More security stuff. Kill double slashes;
        $current =~ s/$/\//g;
        $current =~ s/(\/\/)/\//g;

        # Now we are sure that, there is only one slash per directory and
        # so we can split it using the slash.
        my @dirs = split "/", $current;

        # Just another helper that must be initialized.
        my $headline_path = "";

        # Path fixes. If you do not run the script as:
        # http://yourdomain.com/main.pl, but as http://filelister.yourdomain.com/
        # it happens, that the URLs are not built correctly. Mozilla doesn't mind
        # IE doesn't mind, but at least lynx does!
        unless(url() =~ /main.pl/) {
            $path_line = a({href=>"/"}, "H O M E ");
        } else {
            $path_line = a({href=>"/"}, "H O M E ");
        }

        # Now we use our helper variable headline_path. This stuff builds the nice
        # navigation, which let's you jump back to a higher directory of your choice
        foreach my $dir (@dirs) {
            $headline_path .= "/".$dir unless $dir eq "";
            unless(url() =~ /main.pl/) {
                $path_line .= a({href=>"/?current=$headline_path/"},"$dir")." / ";
            } else {
                $path_line .= a({href=>"?current=$headline_path/"},"$dir")." / ";
            }
        }

        # Loop through the filelist.
        foreach (sort @files) {
            # This is where the hide_files stuff comes into the play
            # Don't ask, why we called the variable ufile. It's a long story...
            my $ufile = 0;
            foreach my $ufile_regex (@$hide_files) {
                if($_ =~ $ufile_regex) { $ufile = 1; }
            }

            # If one of the regex matched, get the next entry - we don't want to
            # show this file....
            next if $ufile;

            # If we are on our root, we don't want to show the 'One up' link,
            # so get back to work and get another file.
            if($current eq '/') {
                next if $_ eq '..';
            }

            # The current directory "." is ignored as well.
            next if $_ eq ".";

            # Check if the "file" it is a directory
            if(-d "$config->{'path_and_url_stuff'}->{'path'}/$current$_") {
                # Is the directory the parent directory? If yes, we want to show the 'One up'-link
                if($_ eq '..') {
                    my $oneup = $current;

                    # Fix the url
                    $oneup =~ s/(.*)\/(?!$)//;
                    $oneup = $1;

                    # Once again, the url fixing stuff... we had this allready....
                    unless(url() =~ /main.pl/) {
                        $oneup_url = "/?current=$oneup/";
                    } else {
                        $oneup_url = "?current=$oneup/";
                    }

                # It is not the parent directory ("..") - fine.
                # It is not the current directory (".") - fine.
                # But we are still in the directory part of this if
                } else {
                    # This is a bit tricky... It may not work on some systems...
                    # If it doesn't for you, cou can - for example - comment it out and set size to: "Unknown".
                    open(FILE, "du -sc $config->{'path_and_url_stuff'}->{'path'}/$current$_|");

                    # The first part of the output from du (at least on my RH systems) is the size.
                    # The second part is the name of the directory - which is ignored - we know the name
                    # allready.
                    my($size) = split(" ", <FILE>);

                    # We can close the pipe now... This isn't very nice in this way, contributions are welcome.
                    close(FILE);

                    # We got the size in bytes from "du".
                    # Now check how big it is and use "MB" and "GB" instead of very long,
                    # not very good readable numbers.
                    # Note to myself: Add KB as well. :-)
                    if ($size <= 1000) { $size = "&lt; 1 MB" }
                    elsif ($size < 1000000) {
                        $size = int($size/1000);
                        $size = $size." MB";
                    } else {
                        $size = int($size/1000000);
                        $size = $size." GB";
                    }

                    # Now create a hash, that we can push into our @dir_list-array.
                    # @dir_list is then finally given to the template (using our main_data-hash).
                    my %dir;

                    # Fill the information we gathered into the hash.
                    $dir{name} = $_;
                    $dir{icon} = $config->{'style_stuff'}->{'icons'}->{'diricon'};
                    unless(url() =~ /main.pl/) {
                        $dir{url}  = "/?current=$current$_/";
                    } else {
                        $dir{url}  = "?current=$current$_/";
                    }
                    $dir{size} = $size;

                    my $dirname = "$config->{'path_and_url_stuff'}->{'path'}/$current$_";

                    # These few lines get the UID of a file and use getpwuid to convert it to a 'name'
                    # You can enable/disable this feature in FileLister.yaml.conf
                    if($config->{'style_stuff'}->{'Show_Owner_Directory'}) {
                        $dir{uid} = (stat($dirname))[4];
                        unless($config->{'style_stuff'}->{'Show_Owner_as_UID_Directory'}) {
                            my ($name,$passwd,$uid,$gid,$quota,$comment,$gcos,$dir,$shell,$expire) = getpwuid($dir{uid});
                            $dir{uid} = $name;
                        }
                    }

                    # These few lines get the GID of a file and use getpwuid to convert it to a 'name'
                    # You can enable/disable this feature in FileLister.yaml.conf
                    if($config->{'style_stuff'}->{'Show_Group_Directory'}) {
                        $dir{gid} = (stat($dirname))[5];
                        unless($config->{'style_stuff'}->{'Show_Group_as_GID_Directory'}) {
                            my ($name,$passwd,$gid,$members) = getgrgid($dir{gid});
                            $dir{gid} = $name;
                        }
                    }

                    # And finally push it into the array
                    push @dir_list, \%dir;
                }

            # At this point we know, that we have done directory stuff - it must be a file.
            # We don't check for links yet and it has never been tested... Do it yourselve if you do have to.
            # But I expect that you get problems with FTP-servers as well, if you use links and therefor
            # I will not code it.... I use FTP and HTTP as download-method...
            } else {
                # Create the currenticon helper-variable.
                my $currenticon = $config->{'style_stuff'}->{'icons'}->{'fileicon'};

                # This stuff is currently very silly and will be changed in future.
                # If we would have something like images/icons/fileicons and our files
                # named <extension>.gif, we could loop through that filelist, we get rid of that...
                # Who want's to code it? :-P
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'zipicon'}   if $_ =~ /\.zip$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'raricon'}   if $_ =~ /\.rar$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'tgzicon'}   if $_ =~ /\.tgz$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'tgzicon'}   if $_ =~ /\.tar.gz$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'gzicon'}    if $_ =~ /\.gz$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'taricon'}   if $_ =~ /\.tar$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'rpmicon'}   if $_ =~ /\.rpm$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'pdficon'}   if $_ =~ /\.pdf$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'patchicon'} if $_ =~ /\.patch$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'patchicon'} if $_ =~ /\.patch.gz$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'shicon'}    if $_ =~ /\.sh$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'plicon'}    if $_ =~ /\.pl$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'texticon'}  if $_ =~ /\.text$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'texticon'}  if $_ =~ /\.txt$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'texicon'}   if $_ =~ /\.tex$/;
                $currenticon = $config->{'style_stuff'}->{'icons'}->{'isoicon'}   if $_ =~ /\.iso$/;

                # Set a helper $filename. Easier to use with the next stuff...
                my $filename="$config->{'path_and_url_stuff'}->{'path'}/$current$_";

                # Get the size of the file.
                # See perldoc -f stat. a snipplet from mine:
                #--------------------------------------------------------------
                #   0 dev      device number of filesystem
                #   1 ino      inode number
                #   2 mode     file mode  (type and permissions)
                #   3 nlink    number of (hard) links to the file
                #   4 uid      numeric user ID of file's owner
                #   5 gid      numeric group ID of file's owner
                #   6 rdev     the device identifier (special files only)
                #   7 size     total size of file, in bytes
                #   8 atime    last access time in seconds since the epoch
                #   9 mtime    last modify time in seconds since the epoch
                #  10 ctime    inode change time in seconds since the epoch (*)
                #  11 blksize  preferred block size for file system I/O
                #  12 blocks   actual number of blocks allocated
                #--------------------------------------------------------------

                my ($size) = (stat($filename))[7];

                # Check for the size and use a similar function (ahhh - why is this no function???) as we did
                # for the directories.
                if ($size <= 1000) { $size = "< 1 KB" }
                elsif ($size < 1000000) {
                    $size = int($size/1000);
                    $size = $size." KB";
                } else {
                    $size = int($size/1000000);
                    $size = $size." MB";
                }

                # Get the 'last modify time' - see the above table...
                my ($sec,     $min,   $hour,   $mday,   $mon,   $year,   $wday,   $yday,   $isdst) = localtime(((stat($filename)))[9]);

                # Get the current time.
                my ($c_sec, $c_min, $c_hour, $c_mday, $c_mon, $c_year, $c_wday, $c_yday, $c_isdst) = localtime(time);

                # Fix the stuff that is returned by localtime...
                $mon++; $c_mon++;
                $year += 1900; $c_year += 1900;

                # Create a hash %file, that is used the same way we used the %dir hash.
                my %file;

                # Fill the information we gathered into the hash.
                $file{name}      = $_;
                $file{icon}      = $currenticon;
                if($config->{'path_and_url_stuff'}->{'enable-ftp-download'}) {
                    $file{url_ftp}   = a({-href=>"$config->{'path_and_url_stuff'}->{'ftp_url'}$current$_"}, "FTP");
                }
                if($config->{'path_and_url_stuff'}->{'enable-http-download'}) {
                    $file{url_http}  = a({-href=>"$config->{'path_and_url_stuff'}->{'http_url'}$current$_"}, "HTTP");
                }
                $file{size}      = $size;
                $file{date}      = sprintf("%4s-%02s-%02s %02s:%02s:%02s", $year, $mon, $mday, $hour, $min, $sec);

                if ($_ =~ /\.rpm$/) {
                    $file{rpminfo} = a({-href=>"/rpminfo.pl"."?file=$current$_",-target=>'_new'}, "Info");
                } else {
                    $file{rpminfo} = '&nbsp;';
                }

                if (-f "$filename.txt") {
                    $file{note} = a({-href=>"$config->{'path_and_url_stuff'}->{'http_url'}$current$_.txt",-target=>'_new'}, "Notes!");
                } else {
                    $file{note} = '&nbsp;';
                }

                # Set default value: ''
                $file{new} = '';

                unless(is_old(time_delta($filename), $config->{'style_stuff'}->{'New_is_max_old'})) {
                    $file{new} = "$config->{'style_stuff'}->{'icons'}->{'new_flash'}";
                }

                
                # These few lines get the UID of a file and use getpwuid to convert it to a 'name'
                # You can enable/disable this feature in FileLister.yaml.conf
                if($config->{'style_stuff'}->{'Show_Owner'}) {
                    $file{uid} = (stat($filename))[4];
                    unless($config->{'style_stuff'}->{Show_Owner_as_UID}) {
                        my ($name,$passwd,$uid,$gid,$quota,$comment,$gcos,$dir,$shell,$expire) = getpwuid($file{uid});
                        $file{uid} = $name;
                    }
                }

                # These few lines get the GID of a file and use getpwuid to convert it to a 'name'
                # You can enable/disable this feature in FileLister.yaml.conf
                if($config->{'style_stuff'}->{'Show_Group'}) {
                    $file{gid} = (stat($filename))[5];
                    unless($config->{'style_stuff'}->{Show_Group_as_GID}) {
                        my ($name,$passwd,$gid,$members) = getgrgid($file{gid});
                        $file{gid} = $name;
                    }
                }

                # Finally push it into the @file_list-array.
                push @file_list, \%file;
            }  
        }
    }

    # This is beta code. We recommend not to use it.
    # If you like it and it work's for you - feel free to use it, but don't blame if, if something goes wrong. :-P
    # And since this is beta code, I will NOT document it!!!
    # And it was an awfull idea. Takes a lot of resources and is no longer used since, there
    # is the search box...
    my @subfolders;
    if($config->{'style_stuff'}->{'SUBFOLDER_ENABLE'}) {
        my $treepath = $config->{'path_and_url_stuff'}->{'path'}.$current;
        open(FILE, "find $treepath -type d | sed -e s^$config->{'path_and_url_stuff'}->{'path'}^^ | sed -e s%^/%% | ");
        my $null = <FILE>;
        while (<FILE>) {
            my $ufile = 0;
            foreach my $ufile_regex (@$hide_files) {
                if($_ =~ $ufile_regex) { $ufile = 1; }
            }

            next if $ufile;
            my $chomped_val = $_;
            chomp($chomped_val);
            unless(url() =~ /main.pl/) {
              push @subfolders, a({href=>url()."/?current=/$chomped_val"}, "/$chomped_val");
            } else {
              push @subfolders, a({href=>url()."?current=/$chomped_val"}, "/$chomped_val");
            }
        }
    }

    # Finally create a hash with all the data, we want to display...
    %$main_data = ( title             => "FileLister $VERSION \@ ".url(),
                    version           => $VERSION,
                    stylesheet        => $config->{'style_stuff'}->{'stylesheet'},
                    SUBFOLDER_ENABLE  => $config->{'style_stuff'}->{'SUBFOLDER_ENABLE'},
                    subfolders        => \@subfolders,
                    bg_color          => $config->{'style_stuff'}->{'FileLister_BGColor'},
                    path_line         => $path_line,
                    oneup_url         => $oneup_url,
                    oneup_text        => 'One up',
                    oneup_icon        => $config->{'style_stuff'}->{'icons'}->{'upicon'},
                    dir_list          => \@dir_list,
                    file_list         => \@file_list,
                    show_user         => $config->{'style_stuff'}->{'Show_Owner'},
                    show_group        => $config->{'style_stuff'}->{'Show_Group'},
                    show_user_dir     => $config->{'style_stuff'}->{'Show_Owner_Directory'},
                    show_group_dir    => $config->{'style_stuff'}->{'Show_Group_Directory'},
                    mailto_on_own     => $config->{'style_stuff'}->{'Automatic_MailTo_HREF_on_Owner'},
                    mailto_domain     => $config->{'style_stuff'}->{'mailto_domain'},
                    use_tooltips      => $config->{'style_stuff'}->{'Use_ToolTips'},
                    use_search        => $config->{'style_stuff'}->{'Use_Search'},
                    search_url        => $config->{'path_and_url_stuff'}->{'search_url'},
                    file_description  => 'Jump to the directory containing this file',
                    ftp_description   => 'Download this file using FTP',
                    http_description  => 'Download this file using HTTP',
                    notes_description => 'Click here to read/download the release notes for this file',
                    info_description  => 'Click here to get more information about this file',
                    comes_from_cache  => $cached);
    }

# Check if caching is enabled. This would be much faster, if we would run mod_perl and
# cache all those things in memory.
if($config->{'caching'}->{'enabled'}) {
    # Check if user wants us to create missing caching directories
    # This should even work on Windows systems and if parent directories do not
    # exist
    unless(-d $config->{'caching'}->{'path'}) {
        if($config->{'caching'}->{'create_missing_cache_dir'}) {
            warn "Createing $config->{'caching'}->{'path'}, because it doesn't exist!" if($config->{'caching'}->{'debug'});
            mkpath($config->{'caching'}->{'path'});
        }
    }

    # Helpers.
    my $store_filename = "FileLister".$current;
    $store_filename =~ s/\/$/$config->{'caching'}->{'suffix'}/g;
    $store_filename =~ s/\//_/g;

    # If caching file for this path does exist
    if(-f "$config->{'caching'}->{'path'}/$store_filename") {
        my $is_expired = is_old(time_delta("$config->{'caching'}->{'path'}/$store_filename"), $config->{'caching'}->{'expire'});

        # Reset stuff from cache, that may has changed, but doesn't need any
        # non-cached data.
        $main_data->{'title'}             = "FileLister $VERSION \@ ".url();
        $main_data->{'version'}           = $VERSION;
        $main_data->{'stylesheet'}        = $config->{'style_stuff'}->{'stylesheet'};
        $main_data->{'bg_color'}          = $config->{'style_stuff'}->{'FileLister_BGColor'};
        $main_data->{'use_search'}        = $config->{'style_stuff'}->{'Use_Search'};
        $main_data->{'search_url'}        = $config->{'path_and_url_stuff'}->{'search_url'};
        $main_data->{'file_description'}  = 'Jump to the directory containing this file';
        $main_data->{'comes_from_cache'}  = $cached;

        # If time-delta (this is configured via configfile) is greater, the cache has expired.
        # Create a new one.
        if($is_expired) {
            store $main_data, "$config->{'caching'}->{'path'}/$store_filename";
            warn "Creating cache for $current, because cache expired" if $config->{'caching'}->{'debug'};
        # Cache not expired.
        } else {
            warn "Cache not yet expired..." if $config->{'caching'}->{'debug'};
        }
    # Caching is enabled, but cache file has not yet be created -> create one.
    } else {
        store $main_data, "$config->{'caching'}->{'path'}/$store_filename";
        warn "Creating cache for $current, because cache doesn't exist" if $config->{'caching'}->{'debug'};
    }
}

# Hand it over to Template::Toolkit - I love that thingy. :-)
$tt->process('templates/main.tpl', $main_data) || die $tt->error;
